package com.qmth.ops.api.controller.export;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.qmth.boot.api.annotation.Aac;
import com.qmth.boot.api.annotation.BOOL;
import com.qmth.boot.core.exception.ParameterException;
import com.qmth.boot.core.exception.UnauthorizedException;
import com.qmth.ops.api.binder.FileFormatBinder;
import com.qmth.ops.api.constants.OpsApiConstants;
import com.qmth.ops.biz.domain.*;
import com.qmth.ops.biz.service.*;
import com.qmth.ops.biz.utils.PropertyFileUtil;
import com.qmth.ops.biz.utils.VersionNumber;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.*;

import javax.annotation.Resource;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.List;
import java.util.stream.Collectors;

@RestController
@RequestMapping(OpsApiConstants.EXPORT_URI_PREFIX + "/config")
@Aac(auth = BOOL.FALSE)
public class ConfigExportController {

    @Resource
    private UserService userService;

    @Resource
    private AppService appService;

    @Resource
    private ModuleService moduleService;

    @Resource
    private EnvService envService;

    @Resource
    private VersionService versionService;

    @Resource
    private ConfigService configService;

    @InitBinder
    public void initBinder(WebDataBinder dataBinder) {
        dataBinder.addCustomFormatter(new FileFormatBinder());
    }

    @RequestMapping("/{appCode}/{moduleCode}/{envCode}/{version}.{format}")
    public void exportVersionConfigFile(@PathVariable String appCode, @PathVariable String moduleCode,
            @PathVariable String envCode, @PathVariable String version, @PathVariable FileFormat format,
            @RequestParam String secret, HttpServletResponse response) throws IOException {
        exportConfigFile(appCode, moduleCode, envCode, version, secret, format, response);
    }

    @RequestMapping("/{appCode}/{moduleCode}/{envCode}.{format}")
    public void exportMasterVersionConfigFile(@PathVariable String appCode, @PathVariable String moduleCode,
            @PathVariable String envCode, @PathVariable FileFormat format, @RequestParam String secret,
            HttpServletResponse response) throws IOException {
        exportConfigFile(appCode, moduleCode, envCode, null, secret, format, response);
    }

    private void exportConfigFile(String appCode, String moduleCode, String envCode, String versionNumber,
            String exportSecret, FileFormat format, HttpServletResponse response) throws IOException {
        User user = userService.findByExportSecret(exportSecret);
        if (user == null || !user.hasRole(Role.OPS)) {
            throw new UnauthorizedException("鉴权失败");
        }
        App app = appService.findByCode(appCode);
        if (app == null) {
            throw new ParameterException("app.code不存在");
        }
        Module module = moduleService.findByAppAndCode(app.getId(), moduleCode);
        if (module == null) {
            throw new ParameterException("module.code不存在");
        }
        Env env = envService.findByAppAndCode(app.getId(), envCode);
        if (env == null) {
            throw new ParameterException("env.code不存在");
        }
        Version version = versionNumber != null ?
                versionService.findByAppAndNumber(app.getId(),
                        new VersionNumber(versionNumber.replaceAll("-", ".").replaceAll("_", "."))) :
                versionService.getById(app.getMasterVersionId());
        if (version == null) {
            throw new ParameterException("version不存在");
        }
        List<ConfigItem> list = configService
                .mergeConfigList(app.getId(), version.getId(), module.getId(), env.getId());
        if (format == FileFormat.PROPERTY) {
            response.reset();
            response.setContentType("application/octet-stream; charset=unicode");
            response.setHeader("Content-Disposition", "attachment; filename=application." + format.getExtension());
            PropertyFileUtil.write(list, response.getOutputStream());
        } else if (format == FileFormat.JSON) {
            response.reset();
            response.setContentType("application/octet-stream; charset=utf-8");
            response.setHeader("Content-Disposition", "attachment; filename=application." + format.getExtension());
            new ObjectMapper().writeValue(response.getOutputStream(),
                    list.stream().collect(Collectors.toMap(ConfigItem::getKey, ConfigItem::getValue)));
        }
    }

}
